/* Bendable Color ePaper Display Prototype
 * ----------------------------------------
 * All wiring diagrams, detailed building instructions, part lists and 3D models of the prototype are avaialable on our project website:
 * https://imld.de/epaper/
 * 
 * If you would like to cite our work in your research, please cite our UIST'21 publication:
 * 
 * Bendable Color ePaper Displays for Novel Wearable Applications and Mobile Visualization.
 * Konstantin Klamka, and Raimund Dachselt.
 * In the 34st Annual ACM Symposium on User Interface Software and Technology Adjunct Proceedings. 
 * UIST '21, Virtual Event. ACM,2021. 
 * DOI: https://doi.org/10.1145/3474349.3480213
 * 
 * Please note that this project is at an early stage and some functions are not yet fully implemented.
 * 
 * License notice
 * Copyright (c) Technische Universität Dresden.
 * 
 * THE SOFTWARE IS PROVIDED AS IS, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// ----------------------------------------
// Includes
// ----------------------------------------

// Arduino
#include <Arduino.h>                      // Arduino

// Bluetooth
#include <bluefruit.h>                    // nRF52840 BLE functions

// Plastic Logic, EPD
#include <PL_smallEPD.h>                  // Generic library for UC8156-based EPDs from Plastic Logic
#include "PL_smallLegio.h"                // Display-specific library for 4pigment media (Legio, 6colors)

// Adafruit, Sensors
#include <Adafruit_LSM6DS33.h>            // LSM6DS33 Motion Sensor

// StrapDisplay, Input
#include "StrapDisplayInputManager.h"     // Input Manager
#include "strapDisplayANCSmsg.h"          // Apple Notification Center Service

// StrapDisplay, Apps
#include "StrapDisplayAppPedometer.h"     // Pedometer App
#include "StrapDisplayAppMediaControls.h" // Media Controls App
#include "StrapDisplayAppDrawer.h"        // App Drawer
#include "StrapDisplayAppWeather.h"       // Weather App
#include "StrapDisplayAppMsgCentral.h"    // Message Central App
#include "StrapDisplayAppPattern.h"       // Pattern App
#include "StrapDisplayAppWorkWear.h"      // Work Wear App
#include "StrapDisplayAppWatchStrap.h"    // Watch Strap App
#include "StrapDisplayAppInfoViz.h"       // InfoVis App

// ----------------------------------------
// BLE function signatures
// ----------------------------------------

// Bluetooth Callback Functions
void disconnect_callback(uint16_t conn_handle, uint8_t reason);
void ancs_notification_callback(AncsNotification_t* notif);
void connect_callback(uint16_t conn_handle);
void connection_secured_callback(uint16_t conn_handle);

// ----------------------------------------
// Defines
// ----------------------------------------

// PIN Definition
#define EPD_RST         12
#define EPD_BUSY        9
#define EPD_CS          5
#define SOFTPOT_LEFT    A0
#define SOFTPOT_MIDDLE  A2
#define SOFTPOT_RIGHT   A1

// ----------------------------------------
// Global Vars
// ----------------------------------------

// > StrapDisplays Applications
StrapDisplayApp               *myCurrentApp;
StrapDisplayAppPedometer      myStrapDisplayAppPedometer;
StrapDisplayAppMediaControls  myStrapDisplayAppMediaControls;
StrapDisplayAppDrawer         myStrapDisplayAppDrawer;
StrapDisplayAppWeather        myStrapDisplayAppWeather;
StrapDisplayAppPattern        myStrapDisplayAppPattern;
StrapDisplayAppWorkWear       myStrapDisplayAppWorkWear;
StrapDisplayAppWatchStrap     myStrapDisplayAppWatchStrap;
StrapDisplayAppInfoViz        myStrapDisplayAppInfoViz;
StrapDisplayAppMsgCentral     myStrapDisplayAppMsgCentral;

// > StrapDisplayInputManager
StrapDisplayInputManager      myStrapDisplayInputManager;

// > Sensor Objects
Adafruit_LSM6DS33 lsm6ds33;

// > ANCS, Bluetooth
#define       BUFSIZE   128
BLEClientDis  bleClientDis;
BLEAncs       bleancs;
char          buffer[BUFSIZE];
const char*   EVENT_STR[] = { "Added", "Modified", "Removed" };
const char*   CAT_STR  [] = { "Other", "Incoming Call", "Missed Call", "Voice Mail", "Social", "Schedule", "Email", "News", "Health and Fitness", "Business and Finance", "Location", "Entertainment"};

// > HID, Bluetooth
BLEDis bledis;
BLEHidAdafruit blehid;

// > EPD
PL_smallLegio epd(EPD_CS, EPD_RST, EPD_BUSY);

// ----------------------------------------
// Setup
// ----------------------------------------
void setup() {

  // Configure Analog Pins as Input
  pinMode(SOFTPOT_LEFT, INPUT);
  pinMode(SOFTPOT_MIDDLE, INPUT);
  pinMode(SOFTPOT_RIGHT, INPUT);

  // Serial Monitor
  Serial.begin(57600);

  // SPI initialisation, UC8156 supp. max 10Mhz (writing) & 6.6Mhz (reading)
  SPI.begin();
  delay(1000);
  SPI.beginTransaction(SPISettings(6000000, MSBFIRST, SPI_MODE0));
  delay(1000);

  // EPD Begin
  epd.begin();

  // Bluetooth: HID Consumer Media Keys + Apple ANCS
  Bluefruit.begin();
  Bluefruit.setName("StrapDisplays");
  Bluefruit.setTxPower(4);
  Bluefruit.Periph.setConnectCallback(connect_callback);
  Bluefruit.Periph.setDisconnectCallback(disconnect_callback);
  Bluefruit.Security.setSecuredCallback(connection_secured_callback);
  bleClientDis.begin();
  Bluefruit.Advertising.restartOnDisconnect(true);
  Bluefruit.Advertising.addService(blehid);
  bleancs.begin();
  bleancs.setNotificationCallback(ancs_notification_callback);
  Bluefruit.Advertising.setInterval(32, 244);
  Bluefruit.Advertising.setFastTimeout(30);
  Bluefruit.Advertising.start(0);
  blehid.begin();
  Bluefruit.Advertising.addFlags(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);
  Bluefruit.Advertising.addTxPower();
  Bluefruit.Advertising.addService(bleancs);
  Bluefruit.ScanResponse.addName();
  Bluefruit.Advertising.start(0);

  // StepCounter
  if (!lsm6ds33.begin_I2C()) {
    Serial.println("Failed to find LSM6DS33 chip");
    while (1) {delay(10);}
  }
  lsm6ds33.setAccelRange(LSM6DS_ACCEL_RANGE_2_G);
  lsm6ds33.setGyroRange(LSM6DS_GYRO_RANGE_250_DPS);
  lsm6ds33.setAccelDataRate(LSM6DS_RATE_26_HZ);
  lsm6ds33.setGyroDataRate(LSM6DS_RATE_26_HZ);
  lsm6ds33.enablePedometer(true);

  myStrapDisplayInputManager.debugSoftPots(false);

  // Setup Pedometer App
  myStrapDisplayAppPedometer.registerEPD(&epd);
  myStrapDisplayAppPedometer.registerLSM6DS33(&lsm6ds33);
  myStrapDisplayAppPedometer.registerStrapDisplayCore(&myStrapDisplayInputManager);
  myStrapDisplayAppPedometer.init();
  delay(50);

  // Setup Media Control App
  myStrapDisplayAppMediaControls.registerEPD(&epd);
  myStrapDisplayAppMediaControls.registerStrapDisplayCore(&myStrapDisplayInputManager);
  myStrapDisplayAppMediaControls.registerBLEHidAdafruit(&blehid);
  myStrapDisplayAppMediaControls.init();
  delay(50);

  // Setup Weather App
  myStrapDisplayAppWeather.registerEPD(&epd);
  // myStrapDisplayAppWeather.setApplicationID(APP_WEATHER);
  myStrapDisplayAppWeather.registerStrapDisplayCore(&myStrapDisplayInputManager);
  myStrapDisplayAppWeather.init();
  delay(50);

  // Setup Message Central App
  myStrapDisplayAppMsgCentral.registerEPD(&epd);
  myStrapDisplayAppMsgCentral.registerStrapDisplayCore(&myStrapDisplayInputManager);
  myStrapDisplayAppMsgCentral.init();
  delay(50);

  // Setup Pattern App
  myStrapDisplayAppPattern.registerEPD(&epd);
  myStrapDisplayAppPattern.registerStrapDisplayCore(&myStrapDisplayInputManager);
  myStrapDisplayAppPattern.init();
  delay(50);

  // Setup Work Wear App
  myStrapDisplayAppWorkWear.registerEPD(&epd);
  myStrapDisplayAppWorkWear.registerStrapDisplayCore(&myStrapDisplayInputManager);
  myStrapDisplayAppWorkWear.init();
  delay(50);

  // Setup Watch+Strap App
  myStrapDisplayAppWatchStrap.registerEPD(&epd);
  myStrapDisplayAppWatchStrap.registerStrapDisplayCore(&myStrapDisplayInputManager);
  myStrapDisplayAppWatchStrap.init();
  delay(50);

  // Setup InfoVis App
  myStrapDisplayAppInfoViz.registerEPD(&epd);
  myStrapDisplayAppInfoViz.registerStrapDisplayCore(&myStrapDisplayInputManager);
  myStrapDisplayAppInfoViz.init();
  delay(50);

  // Setup App Drawer
  myStrapDisplayAppDrawer.registerEPD(&epd);
  myStrapDisplayAppDrawer.registerStrapDisplayApp(&myStrapDisplayAppWeather);
  myStrapDisplayAppDrawer.registerStrapDisplayApp(&myStrapDisplayAppMediaControls);
  myStrapDisplayAppDrawer.registerStrapDisplayApp(&myStrapDisplayAppPedometer);
  myStrapDisplayAppDrawer.registerStrapDisplayApp(&myStrapDisplayAppMsgCentral);
  myStrapDisplayAppDrawer.registerStrapDisplayApp(&myStrapDisplayAppPattern);
  myStrapDisplayAppDrawer.registerStrapDisplayApp(&myStrapDisplayAppWorkWear);
  myStrapDisplayAppDrawer.registerStrapDisplayApp(&myStrapDisplayAppWatchStrap);
  myStrapDisplayAppDrawer.registerStrapDisplayApp(&myStrapDisplayAppInfoViz);
  myStrapDisplayAppDrawer.registerStrapDisplayCore(&myStrapDisplayInputManager);
  myStrapDisplayAppDrawer.init();
  delay(50);

  // Start App Drawer Application
  myCurrentApp = &myStrapDisplayAppDrawer;
  myCurrentApp->show();

  // Register Applications IDs
  myStrapDisplayAppPedometer.setApplicationID(APP_PEDOMETER);
  myStrapDisplayAppMediaControls.setApplicationID(APP_MEDIACONTROLS);
  myStrapDisplayAppWeather.setApplicationID(APP_WEATHER);
  myStrapDisplayAppMsgCentral.setApplicationID(APP_MSG_CENTRAL);
  myStrapDisplayAppPattern.setApplicationID(APP_AESTHETIC);
  myStrapDisplayAppWorkWear.setApplicationID(APP_WORKWEAR);
  myStrapDisplayAppWatchStrap.setApplicationID(APP_WATCHSTRAP);
  myStrapDisplayAppInfoViz.setApplicationID(APP_INFOVIS);
}

// -----------------------------------------
// Main Loop
// -----------------------------------------
void loop() {

  switch (myCurrentApp->runApplication()) {

    // App runs
    case APP_IS_RUNNING:
      break;

    // App terminated
    case APP_EXIT:
      myCurrentApp = &myStrapDisplayAppDrawer;
      myCurrentApp->show();
      break;

    // App change requested to Pedometer
    case APP_PEDOMETER:
      myCurrentApp = &myStrapDisplayAppPedometer;
      myCurrentApp->show();
      break;

    // App change requested to Weather
    case APP_WEATHER:
      myCurrentApp = &myStrapDisplayAppWeather;
      myCurrentApp->show();
      break;

    // App change requested to Media Controls
    case APP_MEDIACONTROLS:
      myCurrentApp = &myStrapDisplayAppMediaControls;
      myCurrentApp->show();
      break;

    // App change requested to Message Central
    case APP_MSG_CENTRAL:
      myCurrentApp = &myStrapDisplayAppMsgCentral;
      myCurrentApp->show();
      break;

    // App change requested to Pattern
    case APP_AESTHETIC:
      myCurrentApp = &myStrapDisplayAppPattern;
      myCurrentApp->show();
      break;

    // App change requested to Work Wear
    case APP_WORKWEAR:
      myCurrentApp = &myStrapDisplayAppWorkWear;
      myCurrentApp->show();
      break;

    // App change requested to Watch+Strap
    case APP_WATCHSTRAP:
      myCurrentApp = &myStrapDisplayAppWatchStrap;
      myCurrentApp->show();
      break;

    // App change requested to InfoViz
    case APP_INFOVIS:
      myCurrentApp = &myStrapDisplayAppInfoViz;
      myCurrentApp->show();
      break;
  }

  delay(1);  // Delay in between for stability
}

// ----------------------------------------
// Bluetooth Callback Functions
// ----------------------------------------

void connect_callback(uint16_t conn_handle) {
  BLEConnection* conn = Bluefruit.Connection(conn_handle);
  bleClientDis.discover(conn_handle);
  if (bleancs.discover(conn_handle)) {
    conn->requestPairing();
  }
}

void connection_secured_callback(uint16_t conn_handle) {
  BLEConnection* conn = Bluefruit.Connection(conn_handle);
  if (!conn->secured()) {
    conn->requestPairing();
  }
  else {
    if (bleancs.discovered()) {
      bleancs.enableNotification();
    }
  }
}

void ancs_notification_callback(AncsNotification_t* notif) {

  StrapDisplayANCSmessage myStrapDisplayANCSmessage;

  delay(250);
  Serial.println("-------");
  uint32_t const uid = notif->uid;

  // Application ID & Name
  char appID[128] = { 0 };
  bleancs.getAppID(uid, appID, sizeof(appID));

  memset(buffer, 0, BUFSIZE);
  bleancs.getAppAttribute(appID, ANCS_APP_ATTR_DISPLAY_NAME, buffer, BUFSIZE);

  Serial.print("AppID: ");
  Serial.println(appID);
  delay(25);
  myStrapDisplayANCSmessage.setANCSAppID(appID);

  String appName = buffer;
  Serial.print("AppName: ");
  Serial.println(appName);
  myStrapDisplayANCSmessage.setANCSAppName(appName);
  delay(25);

  // Get notification Title
  // iDevice often include Unicode "Bidirection Text Control" in the Title.
  // Mostly are U+202D as beginning and U+202C as ending. Let's remove them
  memset(buffer, 0, BUFSIZE);
  if (bleancs.getTitle(uid, buffer, BUFSIZE)) {
    char u202D[3] = { 0xE2, 0x80, 0xAD }; // U+202D in UTF-8
    char u202C[3] = { 0xE2, 0x80, 0xAC }; // U+202C in UTF-8

    int len = strlen(buffer);

    if (0 == memcmp(&buffer[len - 3], u202C, 3)) {
      len -= 3;
      buffer[len] = 0; // chop ending U+202C
    }

    if (0 == memcmp(buffer, u202D, 3)) {
      memmove(buffer, buffer + 3, len - 2); // move null-terminator as well
    }
  }
  String eventType = buffer;
  String eventAction = EVENT_STR[notif->eventID];
  Serial.print("EventType: ");
  delay(25);
  Serial.println(eventType);
  myStrapDisplayANCSmessage.setANCSAppEventType(eventType);
  delay(25);

  Serial.print("eventAction: ");
  delay(25);
  Serial.println(eventAction);
  myStrapDisplayANCSmessage.setANCSAppEventAction(eventAction);
  delay(25);

  Serial.print("Message: ");
  delay(25);
  
  // Get notification Message
  memset(buffer, 0, BUFSIZE);
  bleancs.getMessage(uid, buffer, BUFSIZE);
  delay(25);
  String messageX = buffer;
  delay(25);
  Serial.println(messageX);
  myStrapDisplayANCSmessage.setANCSAppMessage(messageX);

  Serial.println("-------");
  if (messageX != "") {myStrapDisplayAppMsgCentral.addMessage(myStrapDisplayANCSmessage);}
}

void disconnect_callback(uint16_t conn_handle, uint8_t reason) {}
